import ch.bildspur.postfx.builder.*;
import ch.bildspur.postfx.pass.*;
import ch.bildspur.postfx.*;
import fastnoise.*;
import peasy.*;


PeasyCam cam;
FastNoise noise;
Particle[] particles = new Particle[5000];
int tailCount = 30;
PGraphics offscreen;
PostFX fx;

color initialStrokeColor;
color particleColor;

float time = 0.0;
PVector noiseOffsetSpeed = new PVector(1.0, 0.0, 0.0);
//PVector noiseOffsetSpeed = new PVector(10.0, 0.0, 0.0);

void setup() {
  //fullScreen(P2D);
  size(1280, 720, P2D);
  offscreen = createGraphics(width, height, P3D);
  
  fx = new PostFX(this);
  cam = new PeasyCam(this, offscreen, 250);
  cam.setMaximumDistance(1000.0);
  //frameRate(30);
  colorMode(HSB, 1.0);
  offscreen.colorMode(HSB, 1.0);
  initialStrokeColor = color(0.6, 0.8, 1.0, 0.5);
  particleColor = color(0.1, 0.8, 1.0);

  offscreen.noSmooth();
  
  offscreen.beginDraw();
  offscreen.perspective(radians(45), float(width)/float(height), 1.0, 1500.0);
  offscreen.endDraw();

  noise = new FastNoise();
  noise.SetNoiseType(FastNoise.NoiseType.Simplex);
  noise.SetFrequency(0.005);
  //noise.SetFrequency(0.01);
  noise.SetFractalOctaves(8);
  noise.SetFractalGain(0.5);
  noise.SetFractalLacunarity(0.01);
  noise.SetCellularDistanceFunction(FastNoise.CellularDistanceFunction.Euclidean);
  noise.SetCellularReturnType(FastNoise.CellularReturnType.Distance2Add);

  for (int i = 0; i < particles.length; i++) {
    particles[i] = new Particle(getNewParticlePosition(), getNewLifeSpan());
  }
}

void draw() {  
  background(0);

  offscreen.beginDraw();
  offscreen.noFill();
  offscreen.blendMode(ADD);
  offscreen.background(0);

  for (int i = 0; i < particles.length; i++) {
    particles[i].simulate();
    particles[i].update();
  }
  offscreen.endDraw();

  time += 0.1;
  blendMode(BLEND);
  image(offscreen, 0, 0);

  blendMode(SCREEN);
  fx.render(offscreen).rgbSplit(100.0).blur(5, 25.0).bloom(0.0, 20, 40).compose();
}

PVector getNewParticlePosition() {
  //return PVector.random3D().normalize().mult(1.0);
  return PVector.random3D().normalize().mult(100.0);
}

float getNewLifeSpan() {
  return random(1.0, 25.0);
}

PVector noise3d(PVector vec) {
  float x = noise.GetNoise((vec.x + (noiseOffsetSpeed.x * time) ), (vec.y + (noiseOffsetSpeed.x * time) ), (vec.z + (noiseOffsetSpeed.x * time)));
  float y = noise.GetNoise((vec.y + (noiseOffsetSpeed.y * time) + 200.0), (vec.z + (noiseOffsetSpeed.y * time) + 515.0), (vec.x + (noiseOffsetSpeed.y * time) + 345.0));
  float z = noise.GetNoise((vec.z + (noiseOffsetSpeed.z * time) + 645.0), (vec.x + (noiseOffsetSpeed.z * time) + 456.0), (vec.y + (noiseOffsetSpeed.z * time) + 105.0));
  return new PVector(x, y, z);
}

PVector CurlNoise(PVector vec) {
  final float e = 0.0009765525;
  final float e2 = 2.0 * e;
  PVector dx = new PVector(e, 0.0, 0.0);
  PVector dy = new PVector(0.0, e, 0.0);
  PVector dz = new PVector(0.0, 0.0, e);

  PVector x0 = noise3d(PVector.sub(vec, dx));
  PVector x1 = noise3d(PVector.add(vec, dx));
  PVector y0 = noise3d(PVector.sub(vec, dy));
  PVector y1 = noise3d(PVector.add(vec, dy));
  PVector z0 = noise3d(PVector.sub(vec, dz));
  PVector z1 = noise3d(PVector.add(vec, dz));

  float x = (y1.z - y0.z) - (z1.y - z0.y);
  float y = (z1.x - z0.x) - (x1.z - x0.z);
  float z = (x1.y - x0.y) - (y1.x - y0.x);

  PVector force = new PVector(x, y, z);
  return force.div(e2).normalize();
}

void resetAll(){
}

class Particle {
  PVector position;
  PVector pposition;
  PVector velocity;
  float lifespan;
  float age;
  color strokeColor = initialStrokeColor;
  ArrayList<PVector> tails = new ArrayList<PVector>();
  Particle(PVector p, float lifespan) {
    this.lifespan = lifespan;
    this.age = 0;
    position = p;
    velocity = new PVector(0, 0, 0);
  }
  void simulate() {
    position.add(CurlNoise(position).mult(2.0)).add(velocity);
    position.normalize().mult(100);

    if (tails.size() > tailCount) tails.remove(0);
    tails.add(position.copy());
    age += 0.1;
    if (age > lifespan) {
      strokeColor = color(hue(strokeColor), saturation(strokeColor), brightness(strokeColor), alpha(strokeColor) - 0.01);
      if (alpha(strokeColor) <= 0) {
        tails.clear();
        this.velocity = new PVector(0, 0, 0);
        this.position = getNewParticlePosition();
        this.lifespan = getNewLifeSpan();
        this.strokeColor = initialStrokeColor;
        this.age = 0;
      }
    }
  }

  void update() {
    offscreen.strokeWeight(1.0);
    offscreen.stroke(strokeColor);
    for (int i = tails.size() - 1; i >= 1; i--) {
      PVector s = tails.get(i);
      PVector e = tails.get(i-1);
      offscreen.line(s.x, s.y, s.z, e.x, e.y, e.z);
    }
    offscreen.strokeWeight(3.0);
    offscreen.stroke(particleColor);
    offscreen.point(position.x, position.y, position.z);
  }
}



void keyPressed() {
  if(key == 'r'){
    resetAll();
  }
}
